"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const cluster_1 = __importStar(require("cluster"));
const events_1 = require("events");
const types_js_1 = require("../../types.js");
const messages_js_1 = require("../messages.js");
const utils_js_1 = require("../utils.js");
const FORK_RETRIES = 5;
class Pool extends events_1.EventEmitter {
    browser;
    maxRetries;
    config;
    workers = [];
    queue = [];
    forcedStop = false;
    failFast;
    get isRunning() {
        return this.workers.length !== this.freeWorkers.length;
    }
    constructor(config, browser) {
        super();
        this.browser = browser;
        this.failFast = config.failFast;
        this.maxRetries = config.maxRetries;
        this.config = config.browsers[browser];
    }
    async init() {
        const poolSize = Math.max(1, this.config.limit ?? 1);
        // TODO Init queue for workers to smooth browser starting load
        this.workers = (await Promise.all(Array.from({ length: poolSize }).map(() => this.forkWorker()))).filter((workerOrError) => workerOrError instanceof cluster_1.Worker);
        if (this.workers.length != poolSize)
            throw new Error(`Can't instantiate workers for ${this.browser} due many errors`);
        this.workers.forEach((worker) => {
            this.exitHandler(worker);
        });
    }
    start(tests) {
        if (this.isRunning)
            return false;
        this.queue = tests.map(({ id, path }) => ({ id, path, retries: 0 }));
        this.process();
        return true;
    }
    stop() {
        if (!this.isRunning) {
            this.emit('stop');
            return;
        }
        this.forcedStop = true;
        this.queue = [];
    }
    process() {
        const worker = this.getFreeWorker();
        const test = this.queue.at(0);
        if (this.queue.length == 0 && this.workers.length === this.freeWorkers.length) {
            this.forcedStop = false;
            this.emit('stop');
            return;
        }
        if (!worker || !test)
            return;
        worker.isRunning = true;
        const { id } = test;
        this.queue.shift();
        this.sendStatus({ id, status: 'running' });
        this.subscribe(worker, test);
        (0, messages_js_1.sendTestMessage)(worker, { type: 'start', payload: test });
        this.process();
    }
    sendStatus(message) {
        this.emit('test', message);
    }
    getFreeWorker() {
        return this.freeWorkers[Math.floor(Math.random() * this.freeWorkers.length)];
    }
    get aliveWorkers() {
        return this.workers.filter((worker) => !worker.exitedAfterDisconnect);
    }
    get freeWorkers() {
        return this.aliveWorkers.filter((worker) => !worker.isRunning);
    }
    async forkWorker(retry = 0) {
        cluster_1.default.setupPrimary({
            args: ['--browser', this.browser, ...process.argv.slice(2)],
        });
        const worker = cluster_1.default.fork();
        const message = await new Promise((resolve) => {
            const readyHandler = (message) => {
                if (!(0, types_js_1.isWorkerMessage)(message))
                    return;
                worker.off('message', readyHandler);
                resolve(message);
            };
            worker.on('message', readyHandler);
        });
        if (message.type != 'error')
            return worker;
        this.gracefullyKill(worker);
        if (retry == FORK_RETRIES)
            return message.payload;
        return this.forkWorker(retry + 1);
    }
    exitHandler(worker) {
        // eslint-disable-next-line @typescript-eslint/no-misused-promises
        worker.once('exit', async () => {
            if (utils_js_1.isShuttingDown.current)
                return;
            const workerOrError = await this.forkWorker();
            if (!(workerOrError instanceof cluster_1.Worker))
                throw new Error(`Can't instantiate worker for ${this.browser} due many errors`);
            this.exitHandler(workerOrError);
            this.workers[this.workers.indexOf(worker)] = workerOrError;
            this.process();
        });
    }
    gracefullyKill(worker) {
        const timeout = setTimeout(() => {
            worker.kill();
        }, 10000);
        worker.on('exit', () => {
            clearTimeout(timeout);
        });
        (0, messages_js_1.sendShutdownMessage)(worker);
    }
    shouldRetry(test) {
        return test.retries < this.maxRetries && !this.forcedStop;
    }
    handleTestResult(worker, test, result) {
        const shouldRetry = result.status == 'failed' && this.shouldRetry(test);
        if (shouldRetry) {
            test.retries += 1;
            this.queue[this.failFast ? 'unshift' : 'push'](test);
        }
        this.sendStatus({ id: test.id, status: shouldRetry ? 'retrying' : result.status, result });
        worker.isRunning = false;
        setImmediate(() => {
            this.process();
        });
    }
    subscribe(worker, test) {
        const subscriptions = [
            (0, messages_js_1.subscribeOnWorker)(worker, 'worker', (message) => {
                if (message.type != 'error')
                    return;
                subscriptions.forEach((unsubscribe) => {
                    unsubscribe();
                });
                this.gracefullyKill(worker);
                this.handleTestResult(worker, test, { status: 'failed', ...message.payload });
            }),
            (0, messages_js_1.subscribeOnWorker)(worker, 'test', (message) => {
                if (message.type != 'end')
                    return;
                subscriptions.forEach((unsubscribe) => {
                    unsubscribe();
                });
                this.handleTestResult(worker, test, message.payload);
            }),
        ];
    }
}
exports.default = Pool;
//# sourceMappingURL=pool.js.map